module CS334

open Parser

type Expr =
| Name of string
| Assignment of string * Expr
| Value of int
| Seq of Expr * Expr

type Env = Map<string,int>

type EvalResult =
| Ran of value: int
| Fail of message: string

let expr, exprImpl = recparser()
(* a value is a multi-digit number expression *)
let value = pmany1 pdigit |>> stringify |>> int |>> Value <!> "value"
(* an alphabetic string *)
let alphastr = pmany1 pletter |>> stringify <!> "alphastr"
(* a name is an alphabetic string expression *)
let name = alphastr |>> Name <!> "name"
let assignSymb = pbetween pws0 pws0 (pstr ":=") <!> "assignSymb"
(* assignment expression looks like: "x := 2" or "x := y := 3" *)
let assignment = pseq (pleft alphastr assignSymb) expr Assignment <!> "assignment"
(* a sequence expression consists of two expressions, separated by a newline *)
let sequence =
    pmany1 ((pleft expr pnl) <|> expr)
    |>> (fun es ->
            es
            |> List.rev
            |> List.reduce (fun acc e -> Seq(e, acc))
        )
    <!> "sequence"
(* an expression is a value, an assignment, a name, or a sequence of any of these *)
exprImpl := value <|> assignment <|> name <!> "expr"
let grammar = pleft sequence peof <!> "grammar"

(* a pretty print function to aid in debugging *)
let rec pp(e: Expr) : string =
    match e with
    | Name n -> n
    | Assignment (n,e) ->
        let s = pp e
        sprintf "%s := %s" n s
    | Value v -> string v
    | Seq (e1, e2) ->
        let s1 = pp e1
        let s2 = pp e2
        sprintf "%s\n%s" s1 s2

let rec eval(e: Expr)(env: Map<string, int>) : int * Env =
    match e with
    | Value v ->
        (* just return the value and unmodified environment *)
        v, env
    | Name n ->
        (* first check to see if the name exists *)
        if Map.containsKey n env then
            (* if so, return value and unmodified environment *)
            Map.find n env, env
        else
            (* otherwise, this is a malformed program; quit *)
            printfn "Unknown variable '%s'" n
            exit 1
    | Assignment (n, e) ->
        (* first evaluate e, then return both the value and the new environment *)
        let v, env' = eval e env
        v, Map.add n v env'
    | Seq (e1,e2) ->
        (* first evaluate e1, discarding result but keeping the new environment,
         * then evalute e2 with the new environment, and return the value of
         * e2 and the new environmetn *)
        let _, env'  = eval e1 env
        eval e2 env'

let run input : EvalResult =
    let i = prepare input
    match grammar i with
    | Success(ast, _) ->
//        printfn "--- START DEBUG---\n%s" (pp ast)
//        printfn "---END DEBUG---"
        let v, _ = eval ast Map.empty
        Ran v
    | Failure(pos, rule) ->
        let msg = sprintf "Cannot parse input at pos %d in rule '%s':" pos rule
        let diag = diagnosticMessage 20 pos input msg
        Fail diag
